# coding: utf8
# Part of the bioread package for reading BIOPAC data.
#
# Copyright (c) 2023 Board of Regents of the University of Wisconsin System
#
# Written Nate Vack <njvack@wisc.edu> with research from John Ollinger
# at the Waisman Laboratory for Brain Imaging and Behavior, University of
# Wisconsin-Madison
# Project home: http://github.com/njvack/bioread

# This is an enumeration of known file revisions, for use in header parsing.

V_ALL = 0
V_20a = 30
V_20b = 31
V_20r = 32
V_207 = 33
V_30r = 34
V_303 = 35
V_35x = 36
V_36x = 37
V_370 = 38
V_373 = 39
V_381 = 41
V_37P = 42
V_382 = 43
V_38P = 44
V_390 = 45
V_400B = 61  # Earliest 'late version' one I've seen so far...
V_400 = 68
V_401 = 76
V_402 = 78  # Unsure about exact versions... but this is Post-Some-Big-Change.
V_41a = 80  # But I'm honestly guessing about some of these.
V_410 = 83
V_411 = 84
V_420 = 108
V_42x = 121  # Not yet sure what version this is
V_430 = 124
V_440 = 128
V_501 = 132
